#include <Kalman.h>
#include "I2Cdev.h"
#include "MPU6050.h"
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include <WiFi.h>
#include <WebServer.h>
#include <Preferences.h>
#include "driver/pcnt.h"

WebServer server(80);

const char ssid[] = "SHISEIGYO-2W";  // SSID
const char pass[] = "password";   // password

const IPAddress ip(192, 168, 22, 22);      // IPアドレス
const IPAddress subnet(255, 255, 255, 0); // サブネットマスク


#define PULSE_INPUT_PIN_X 35
#define PULSE_CTRL_PIN_X 34
#define PWM_pinX 26
#define brakeX 18
#define rote_pinX 5

#define PULSE_INPUT_PIN_Y 39
#define PULSE_CTRL_PIN_Y 36
#define PWM_pinY 25
#define brakeY 17
#define rote_pinY 16

#define CH_X 0
#define CH_Y 1

#define servoL1 27
#define servoL2 14
#define servoR1 4
#define servoR2 13

#define CH_servoL1 14
#define CH_servoL2 15
#define CH_servoR1 12
#define CH_servoR2 13

unsigned long oldTime = 0, loopTime, nowTime;
float dt;

volatile byte posX, posY;
int16_t enc_countX = 0, enc_countY = 0;

int pwmDutyL, DutyIniX = 1020, pwmDutyX, DutyIniY = DutyIniX, pwmDutyY;

int GetUP = 0;

float Mx, My;
float AjY = 0.0, AjX = 0.0;

String modeBtAd = "off";
String modeBtBack = "off";
String modeBtL = "off";
String modeBtR = "off";

//足の長さ
float L1 = 24.0, L2 = 42.0, l = 35.0;

//----------------------------
//調整パラメータ
int Tms=400;
int Y1 = 18, Y2 = 24;
int XX = 4;

float Kp = 7;
float Kd = 5.2;
float Kw = 0.4;
float IDRS = 0.6;

int loopDelay = 3;
//----------------------------


float servoValL1, servoValL2, servoValR1, servoValR2;

//タイマー関連
unsigned long time_ms;
unsigned long time_mSt=0;

int x0 = 0;  

int btn_ad = 0;
int btn_back = 0;
int btn_L = 0;
int btn_R = 0;


MPU6050 mpu;

int16_t ax, ay, az;
int16_t gx, gy, gz;

float accX = 0, accY = 0, accZ = 0;
float gyroX = 0, gyroY = 0, gyroZ = 0;

float theta_X = 0.0, theta_Y = 0.0;
float theta_Xdot = 0.0, theta_Ydot = 0.0;

Kalman kalmanX, kalmanY;
float kalAngleX, kalAngleDotX, kalAngleY, kalAngleDotY;

Preferences preferences;

#define SCREEN_WIDTH 128 // OLED display width, in pixels
#define SCREEN_HEIGHT 64 // OLED display height, in pixels

// Declaration for an SSD1306 display connected to I2C (SDA, SCL pins)
#define OLED_RESET     -1 // Reset pin # (or -1 if sharing Arduino reset pin)
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire1, OLED_RESET);


//加速度センサから傾きデータ取得 [deg]
void get_theta() {
  mpu.getAcceleration(&ax, &ay, &az);
  accX = ax / 16384.0;
  accY = ay / 16384.0;
  accZ = az / 16384.0;
  
  //傾斜角導出 単位はdeg
  theta_X  = atan2(+1.0 * accY , accZ) * 180.0/PI;
  theta_Y  = atan2(-1.0 * accX , accZ) * 180.0/PI;
}

//角速度取得
void get_gyro_data() {
  mpu.getRotation(&gx, &gy, &gz);
  gyroX = gx / 131.072;
  gyroY = gy / 131.072;

  theta_Xdot = gyroX;
  theta_Ydot = gyroY;
}


//ブラウザ表示
void handleRoot() {
  
  String temp ="<!DOCTYPE html> \n<html lang=\"ja\">";
  temp +="<head>";
  temp +="<meta charset=\"utf-8\">";
  temp +="<title>SHISEIGYO-2 Walker</title>";
  temp +="<meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">";
  temp +="<style>";
  temp +=".container{";
  temp +="  max-width: 500px;";
  temp +="  margin: auto;";
  temp +="  text-align: center;";
  temp +="  font-size: 1.2rem;";
  temp +="}";
  temp +="span,.pm{";
  temp +="  display: inline-block;";
  temp +="  border: 1px solid #ccc;";
  temp +="  width: 50px;";
  temp +="  height: 30px;";
  temp +="  vertical-align: middle;";
  temp +="  margin-bottom: 20px;";
  temp +="}";
  temp +="span{";
  temp +="  width: 120px;";
  temp +="}";
  temp +="button{";
  temp +="  width: 100px;";
  temp +="  height: 40px;";
  temp +="  font-weight: bold;";
  temp +="  margin-bottom: 20px;";
  temp +="}";
  temp +="button.on{ background:lime; color:white; }";
  temp +=".column-2{ max-width:250px; margin:auto; text-align:center; display:flex; justify-content:space-between; flex-wrap:wrap; }";
  temp +="</style>";
  temp +="</head>";
  
  temp +="<body>";
  temp +="<div class=\"container\">";
  temp +="<h3>SHISEIGYO-2 Walker</h3>";

  //歩行ボタン
  temp +="<div class=\"column-2\">";
  temp +="<button class=\"" + modeBtAd + "\" type=\"button\" ><a href=\"/ad\">前進</a></button><br>";
  temp +="<button class=\"" + modeBtBack + "\" type=\"button\" ><a href=\"/back\">後進</a></button><br>";
  temp +="<button class=\"" + modeBtL + "\" type=\"button\" ><a href=\"/L\">左旋回</a></button><br>";
  temp +="<button class=\"" + modeBtR + "\" type=\"button\" ><a href=\"/R\">右旋回</a></button><br>";
  temp +="</div>";

  //Tms
  temp +="歩行周期 Tms<br>";
  temp +="<a class=\"pm\" href=\"/TmsM\">-</a>";
  temp +="<span>" + String(Tms) + "</span>";
  temp +="<a class=\"pm\" href=\"/TmsP\">+</a><br>";

  //Y1
  temp +="足上げ低 Y1<br>";
  temp +="<a class=\"pm\" href=\"/Y1M\">-</a>";
  temp +="<span>" + String(Y1) + "</span>";
  temp +="<a class=\"pm\" href=\"/Y1P\">+</a><br>";

  //Y2
  temp +="足上げ高 Y2<br>";
  temp +="<a class=\"pm\" href=\"/Y2M\">-</a>";
  temp +="<span>" + String(Y2) + "</span>";
  temp +="<a class=\"pm\" href=\"/Y2P\">+</a><br>";

  //XX
  temp +="歩幅 XX<br>";
  temp +="<a class=\"pm\" href=\"/XXm\">-</a>";
  temp +="<span>" + String(XX) + "</span>";
  temp +="<a class=\"pm\" href=\"/XXp\">+</a><br>";

  //Kp
  temp +="Kp<br>";
  temp +="<a class=\"pm\" href=\"/KpM\">-</a>";
  temp +="<span>" + String(Kp) + "</span>";
  temp +="<a class=\"pm\" href=\"/KpP\">+</a><br>";

  //Kd
  temp +="Kd<br>";
  temp +="<a class=\"pm\" href=\"/KdM\">-</a>";
  temp +="<span>" + String(Kd) + "</span>";
  temp +="<a class=\"pm\" href=\"/KdP\">+</a><br>";

  //Kw
  temp +="Kw<br>";
  temp +="<a class=\"pm\" href=\"/KwM\">-</a>";
  temp +="<span>" + String(Kw) + "</span>";
  temp +="<a class=\"pm\" href=\"/KwP\">+</a><br>";

  //IDRS
  temp +="IDRS<br>";
  temp +="<a class=\"pm\" href=\"/IDRSm\">-</a>";
  temp +="<span>" + String(IDRS) + "</span>";
  temp +="<a class=\"pm\" href=\"/IDRSp\">+</a><br>";

  //loopDelay
  temp +="loopDelay<br>";
  temp +="<a class=\"pm\" href=\"/loopDelayM\">-</a>";
  temp +="<span>" + String(loopDelay) + "</span>";
  temp +="<a class=\"pm\" href=\"/loopDelayP\">+</a><br>";
   
  temp +="</div>";
  temp +="</body>";
  server.send(200, "text/HTML", temp);

}

void handleAd() {
  if(modeBtAd == "off"){
    modeBtAd = "on";
    modeBtBack = "off";
    modeBtL = "off";
    modeBtR = "off";
    
    btn_ad = 1;
    btn_back = 0;
    btn_L = 0;
    btn_R = 0;
  }else{
    modeBtAd = "off";
    btn_ad = 0;
    btn_back = 0;
    btn_L = 0;
    btn_R = 0;
  }
  handleRoot();
}

void handleBack() {
  if(modeBtBack == "off"){
    modeBtAd = "off";
    modeBtBack = "on";
    modeBtL = "off";
    modeBtR = "off";
    
    btn_ad = 0;
    btn_back = 1;
    btn_L = 0;
    btn_R = 0;
  }else{
    modeBtBack = "off";
    btn_ad = 0;
    btn_back = 0;
    btn_L = 0;
    btn_R = 0;
  }
  handleRoot();
}

void handleL() {
  if(modeBtL == "off"){
    modeBtAd = "off";
    modeBtBack = "off";
    modeBtL = "on";
    modeBtR = "off";
    
    btn_ad = 0;
    btn_back = 0;
    btn_L = 1;
    btn_R = 0;
  }else{
    modeBtL = "off";
    btn_ad = 0;
    btn_back = 0;
    btn_L = 0;
    btn_R = 0;
  }
  handleRoot();
}

void handleR() {
  if(modeBtR == "off"){
    modeBtAd = "off";
    modeBtBack = "off";
    modeBtL = "off";
    modeBtR = "on";
    
    btn_ad = 0;
    btn_back = 0;
    btn_L = 0;
    btn_R = 1;
  }else{
    modeBtR = "off";
    btn_ad = 0;
    btn_back = 0;
    btn_L = 0;
    btn_R = 0;
  }
  handleRoot();
}

void handleTmsM() {
  if(Tms > 100){
    Tms -= 20;
    preferences.putInt("Tms", Tms);
  }
  handleRoot();
}
void handleTmsP() {
  if(Tms <= 1000){
    Tms += 20;
    preferences.putInt("Tms", Tms);
  }
  handleRoot();
}

void handleY1M() {
  if(Y1 > 16){
    Y1 -= 2;
    preferences.putInt("Y1", Y1);
  }
  handleRoot();
}
void handleY1P() {
  if(Y1 <= 40){
    Y1 += 2;
    preferences.putInt("Y1", Y1);
  }
  handleRoot();
}

void handleY2M() {
  if(Y2 > 20){
    Y2 -= 2;
    preferences.putInt("Y2", Y2);
  }
  handleRoot();
}
void handleY2P() {
  if(Y2 <= 50){
    Y2 += 2;
    preferences.putInt("Y2", Y2);
  }
  handleRoot();
}

void handleXXm() {
  if(XX > 0){
    XX -= 1;
    preferences.putInt("XX", XX);
  }
  handleRoot();
}
void handleXXp() {
  if(XX <= 20){
    XX += 1;
    preferences.putInt("XX", XX);
  }
  handleRoot();
}

void handleKpM() {
  if(Kp > 2){
    Kp -= 0.2;
    preferences.putFloat("Kp", Kp);
  }
  handleRoot();
}
void handleKpP() {
  if(Kp <= 10){
    Kp += 0.2;
    preferences.putFloat("Kp", Kp);
  }
  handleRoot();
}

void handleKdM() {
  if(Kd > 2){
    Kd -= 0.2;
    preferences.putFloat("Kd", Kd);
  }
  handleRoot();
}
void handleKdP() {
  if(Kd <= 10){
    Kd += 0.2;
    preferences.putFloat("Kd", Kd);
  }
  handleRoot();
}

void handleKwM() {
  if(Kw > 0){
    Kw -= 0.1;
    preferences.putFloat("Kw", Kw);
  }
  handleRoot();
}
void handleKwP() {
  if(Kw <= 10){
    Kw += 0.1;
    preferences.putFloat("Kw", Kw);
  }
  handleRoot();
}

void handleIDRSm() {
  if(IDRS > 0){
    IDRS -= 0.1;
    preferences.putFloat("IDRS", IDRS);
  }
  handleRoot();
}
void handleIDRSp() {
  if(IDRS <= 10){
    IDRS += 0.1;
    preferences.putFloat("IDRS", IDRS);
  }
  handleRoot();
}

void handleloopDelayM() {
  if(loopDelay > 0){
    loopDelay -= 1;
    preferences.putInt("loopDelay", loopDelay);
  }
  handleRoot();
}
void handleloopDelayP() {
  if(loopDelay <= 10){
    loopDelay += 1;
    preferences.putInt("loopDelay", loopDelay);
  }
  handleRoot();
}


//Core0
void disp(void *pvParameters) {
  Wire1.begin(0, 15); //SDA,SCL
  display.begin(SSD1306_SWITCHCAPVCC, 0x3C);

  // Clear the buffer
  display.clearDisplay();
  
  for (;;){
    server.handleClient();
    disableCore0WDT();


    if(btn_ad == 0 && btn_back == 0 && btn_L == 0 && btn_R == 0){ //停止
      //ディスプレイ
      display.setTextSize(1);             // Normal 1:1 pixel scale
      display.setTextColor(SSD1306_WHITE);        // Draw white text
      
      display.setCursor(0,0);            
      display.println(kalAngleX + AjX, 1);
      display.setCursor(50,0);                        
      display.println(kalAngleY + AjY, 1);
  
      int LineY = map(kalAngleY + AjY, 20, -20, 0, 127);
      if(LineY >= 0 && LineY < 128){
        if(abs(kalAngleY) <= 1.0){
          display.fillRect(LineY-2, 16, 5, display.height()-1, SSD1306_WHITE); //(左上x, 左上y, 幅, 高さ, 線の色
        }else{
          display.drawLine(LineY, 16, LineY, display.height()-1, SSD1306_WHITE); //始点x, 始点y, 終点x, 終点y, 線の色
        }
      }
      
      int LineX = map(kalAngleX + AjX, 8, -8, 16, 63);
      if(LineX >= 16 && LineX < 64){
        if(abs(kalAngleX) <= 1.0){
          display.fillRect(0, LineX - 2, display.width()-1, 5, SSD1306_WHITE); //(左上x, 左上y, 幅, 高さ, 線の色
        }else{
          display.drawLine(0, LineX, display.width()-1, LineX, SSD1306_WHITE);
        }
      }
        
      display.display(); 
      delay(30);
      display.clearDisplay();
      
      x0 = 0;
      ikL(0, Y2);
      ikR(0, Y2);
     
    }else{
      if(btn_ad){ //前進
        float tim,tt;
        
        ms_timer_st();
        tim=0;
        while(tim<Tms){
          tim=ms_timer_rd();
          tt=float(tim*PI/Tms);
          ikL(XX*cos(tt),Y2);
          ikR(-XX*cos(tt),Y2-(Y2-Y1)*sin(tt));
        }

        ms_timer_st();
        tim=0;
        while(tim<Tms){
          tim=ms_timer_rd();
          tt=float(tim*PI/Tms);
          ikL(-XX*cos(tt),Y2-(Y2-Y1)*sin(tt));
          ikR(XX*cos(tt),Y2);
        }
      }
  
      if(btn_back){ //後進
        float tim,tt;

        ms_timer_st();
        tim=0;
        while(tim<Tms){
          tim=ms_timer_rd();
          tt=float(tim*PI/Tms);
          ikL(-XX*cos(tt),Y2);
          ikR(XX*cos(tt),Y2-(Y2-Y1)*sin(tt));
        }

        ms_timer_st();
        tim=0;
        while(tim<Tms){
          tim=ms_timer_rd();
          tt=float(tim*PI/Tms);
          ikL(XX*cos(tt),Y2-(Y2-Y1)*sin(tt));
          ikR(-XX*cos(tt),Y2);
        }
      }
  
      if(btn_L){ //左旋回
        float tim,tt;

        ms_timer_st();
        tim=0;
        while(tim<Tms){
          tim=ms_timer_rd();
          tt=float(tim*PI/Tms);
          ikL(-XX*cos(tt),Y2);
          ikR(0, Y2 - 2);
        }

        ms_timer_st();
        tim=0;
        while(tim<Tms){
          tim=ms_timer_rd();
          tt=float(tim*PI/Tms);
          ikL(XX*cos(tt),Y2-(Y2-Y1)*sin(tt));
          ikR(0, Y2 - 2);
        }
      }
  
      if(btn_R){ //右旋回
        float tim,tt;

        ms_timer_st();
        tim=0;
        while(tim<Tms){
          tim=ms_timer_rd();
          tt=float(tim*PI/Tms);
          ikL(XX*cos(tt),Y2);
          ikR(0, Y2 - 2);
        }

        ms_timer_st();
        tim=0;
        while(tim<Tms){
          tim=ms_timer_rd();
          tt=float(tim*PI/Tms);
          ikL(-XX*cos(tt),Y2-(Y2-Y1)*sin(tt));
          ikR(0, Y2 - 2);
        }
      }
    }
  }
}


void setup() {
  Wire.begin();
  Serial.begin(115200);


  pinMode(brakeX, OUTPUT);
  pinMode(brakeY, OUTPUT);

  pinMode(rote_pinX, OUTPUT);
  pinMode(rote_pinY, OUTPUT);

  digitalWrite(brakeX, LOW);
  digitalWrite(brakeY, LOW);
  

  pcnt_config_t pcnt_config_A_X;   
        pcnt_config_A_X.pulse_gpio_num = PULSE_INPUT_PIN_X;
        pcnt_config_A_X.ctrl_gpio_num = PULSE_CTRL_PIN_X;
        pcnt_config_A_X.lctrl_mode = PCNT_MODE_REVERSE;
        pcnt_config_A_X.hctrl_mode = PCNT_MODE_KEEP;
        pcnt_config_A_X.channel = PCNT_CHANNEL_0;
        pcnt_config_A_X.unit = PCNT_UNIT_0;
        pcnt_config_A_X.pos_mode = PCNT_COUNT_INC;
        pcnt_config_A_X.neg_mode = PCNT_COUNT_DEC;

  pcnt_config_t pcnt_config_B_X;   
        pcnt_config_B_X.pulse_gpio_num = PULSE_CTRL_PIN_X;
        pcnt_config_B_X.ctrl_gpio_num = PULSE_INPUT_PIN_X;
        pcnt_config_B_X.lctrl_mode = PCNT_MODE_KEEP;
        pcnt_config_B_X.hctrl_mode = PCNT_MODE_REVERSE;
        pcnt_config_B_X.channel = PCNT_CHANNEL_1;
        pcnt_config_B_X.unit = PCNT_UNIT_0;
        pcnt_config_B_X.pos_mode = PCNT_COUNT_INC;
        pcnt_config_B_X.neg_mode = PCNT_COUNT_DEC;

  pcnt_unit_config(&pcnt_config_A_X);
  pcnt_unit_config(&pcnt_config_B_X);
  pcnt_counter_pause(PCNT_UNIT_0);
  pcnt_counter_clear(PCNT_UNIT_0);//カウンタ初期化

  pcnt_config_t pcnt_config_A_Y;   
        pcnt_config_A_Y.pulse_gpio_num = PULSE_INPUT_PIN_Y;
        pcnt_config_A_Y.ctrl_gpio_num = PULSE_CTRL_PIN_Y;
        pcnt_config_A_Y.lctrl_mode = PCNT_MODE_REVERSE;
        pcnt_config_A_Y.hctrl_mode = PCNT_MODE_KEEP;
        pcnt_config_A_Y.channel = PCNT_CHANNEL_0;
        pcnt_config_A_Y.unit = PCNT_UNIT_1;
        pcnt_config_A_Y.pos_mode = PCNT_COUNT_INC;
        pcnt_config_A_Y.neg_mode = PCNT_COUNT_DEC;

  pcnt_config_t pcnt_config_B_Y;   
        pcnt_config_B_Y.pulse_gpio_num = PULSE_CTRL_PIN_Y;
        pcnt_config_B_Y.ctrl_gpio_num = PULSE_INPUT_PIN_Y;
        pcnt_config_B_Y.lctrl_mode = PCNT_MODE_KEEP;
        pcnt_config_B_Y.hctrl_mode = PCNT_MODE_REVERSE;
        pcnt_config_B_Y.channel = PCNT_CHANNEL_1;
        pcnt_config_B_Y.unit = PCNT_UNIT_1;
        pcnt_config_B_Y.pos_mode = PCNT_COUNT_INC;
        pcnt_config_B_Y.neg_mode = PCNT_COUNT_DEC;

  pcnt_unit_config(&pcnt_config_A_Y);
  pcnt_unit_config(&pcnt_config_B_Y);
  pcnt_counter_pause(PCNT_UNIT_1);
  pcnt_counter_clear(PCNT_UNIT_1);//カウンタ初期化

  pcnt_counter_resume(PCNT_UNIT_0);//カウンタ初期化
  pcnt_counter_resume(PCNT_UNIT_1);//カウンタ初期化
  
  // initialize device
  Serial.println("Initializing I2C devices...");
  mpu.initialize();

  // verify connection
  Serial.println("Testing device connections...");
  Serial.println(mpu.testConnection() ? "MPU6050 connection successful" : "MPU6050 connection failed");


  mpu.setXAccelOffset(-4562);
  mpu.setYAccelOffset(436);
  mpu.setZAccelOffset(835);
  mpu.setXGyroOffset(39);
  mpu.setYGyroOffset(-17);
  mpu.setZGyroOffset(0);
  
  
  ledcSetup(CH_X, 20000, 10);
  ledcAttachPin(PWM_pinX, CH_X);
  ledcSetup(CH_Y, 20000, 10);
  ledcAttachPin(PWM_pinY, CH_Y);

  ledcSetup(CH_servoL1, 50, 16);
  ledcAttachPin(servoL1, CH_servoL1);
  ledcSetup(CH_servoL2, 50, 16);
  ledcAttachPin(servoL2, CH_servoL2);
  ledcSetup(CH_servoR1, 50, 16);
  ledcAttachPin(servoR1, CH_servoR1);
  ledcSetup(CH_servoR2, 50, 16);
  ledcAttachPin(servoR2, CH_servoR2);

  delay(500); 

  get_theta();
  kalmanX.setAngle(theta_X);
  kalmanY.setAngle(theta_Y);
  

  //パラメータ初期値取得
  preferences.begin("parameter", false);

  Tms = preferences.getInt("Tms", Tms);
  Y1 = preferences.getInt("Y1", Y1);
  Y2 = preferences.getInt("Y2", Y2);
  XX = preferences.getInt("XX", XX);
  Kp = preferences.getFloat("Kp", Kp);
  Kd = preferences.getFloat("Kd", Kd);
  Kw = preferences.getFloat("Kw", Kw);
  IDRS = preferences.getFloat("IDRS", IDRS);
  loopDelay = preferences.getInt("loopDelay", loopDelay);

  
  WiFi.softAP(ssid, pass);           // SSIDとパスの設定
  delay(100);                        // 追記：このdelayを入れないと失敗する場合がある
  WiFi.softAPConfig(ip, ip, subnet); // IPアドレス、ゲートウェイ、サブネットマスクの設定
  
  IPAddress myIP = WiFi.softAPIP();  // WiFi.softAPIP()でWiFi起動

  server.on("/", handleRoot); 
  
  server.on("/ad", handleAd);
  server.on("/back", handleBack);
  server.on("/L", handleL);
  server.on("/R", handleR);

  server.on("/TmsM", handleTmsM);
  server.on("/TmsP", handleTmsP);
  server.on("/Y1M", handleY1M);
  server.on("/Y1P", handleY1P);
  server.on("/Y2M", handleY2M);
  server.on("/Y2P", handleY2P);
  server.on("/XXm", handleXXm);
  server.on("/XXp", handleXXp);

  server.on("/KpM", handleKpM);
  server.on("/KpP", handleKpP);
  server.on("/KdM", handleKdM);
  server.on("/KdP", handleKdP);
  server.on("/KwM", handleKwM);
  server.on("/KwP", handleKwP);
  server.on("/IDRSm", handleIDRSm);
  server.on("/IDRSp", handleIDRSp);
  server.on("/loopDelayM", handleloopDelayM);
  server.on("/loopDelayP", handleloopDelayP);

  server.begin();
  

  //ディスプレイ表示 タスク
  xTaskCreatePinnedToCore(
    disp
    ,  "disp"   // A name just for humans
    ,  4096  // This stack size can be checked & adjusted by reading the Stack Highwater
    ,  NULL
    ,  1  // Priority, with 3 (configMAX_PRIORITIES - 1) being the highest, and 0 being the lowest.
    ,  NULL 
    ,  0);
   
}


void loop() {
  nowTime = micros();
  loopTime = nowTime - oldTime;
  oldTime = nowTime;
  
  dt = (float)loopTime / 1000000.0; //sec
  
  //モータの角速度算出
  pcnt_get_counter_value(PCNT_UNIT_0, &enc_countX);
  float theta_XdotWheel = -1.0 * float(enc_countX) * 0.9 / dt;
  pcnt_counter_clear(PCNT_UNIT_0);

  pcnt_get_counter_value(PCNT_UNIT_1, &enc_countY);
  float theta_YdotWheel = -1.0 * float(enc_countY) * 0.9 / dt;
  pcnt_counter_clear(PCNT_UNIT_1);

  get_theta();
  get_gyro_data();
    
  //カルマンフィルタ 姿勢 傾き
  kalAngleX = kalmanX.getAngle(theta_X, theta_Xdot, dt);
  kalAngleY = kalmanY.getAngle(theta_Y, theta_Ydot, dt);
  
  //カルマンフィルタ 姿勢 角速度
  kalAngleDotX = kalmanX.getRate();
  kalAngleDotY = kalmanY.getRate();

    
  if ((fabs(kalAngleX) < 1 && fabs(kalAngleY) < 1) && GetUP == 0 ){
    GetUP = 99;
  }

  if(GetUP == 99){
    //ブレーキ
    if(fabs(kalAngleX) > 12.0 || fabs(kalAngleY) > 12.0){
      digitalWrite(brakeY, LOW);
      digitalWrite(brakeX, LOW);
      GetUP = 0;
      AjX = 0.0;
      AjY = 0.0;
    }else{
      digitalWrite(brakeX, HIGH);
      digitalWrite(brakeY, HIGH);
    }
     
      
    //モータ回転
    AjX +=  IDRS * theta_XdotWheel / 1000000.0;
    AjY +=  IDRS * theta_YdotWheel / 1000000.0;
     
    Mx = Kp * (kalAngleX + AjX) / 90.0 + Kd * kalAngleDotX / 500.0 + Kw * theta_XdotWheel / 10000.0;
    My = Kp * (kalAngleY + AjY) / 90.0 + Kd * kalAngleDotY / 500.0 + Kw * theta_YdotWheel / 10000.0;
    
    Mx = max(-1.0f, min(1.0f, Mx));
    pwmDutyX = 1023 * (1.0 - fabs(Mx));
    My = max(-1.0f, min(1.0f, My));
    pwmDutyY = 1023 * (1.0 - fabs(My));
  
    //回転方向 
    if(pwmDutyX > 1000){
      digitalWrite(brakeX, LOW);
      ledcWrite(CH_X, 1023);
    }else if(Mx > 0.0){
      digitalWrite(rote_pinX, HIGH);
      ledcWrite(CH_X, pwmDutyX);
    }else{
      digitalWrite(rote_pinX, LOW);
      ledcWrite(CH_X, pwmDutyX);
    }
  
    if(pwmDutyY > 1000){
      digitalWrite(brakeY, LOW);
      ledcWrite(CH_Y, 1023);
    }else if(My < 0.0){
      digitalWrite(rote_pinY, LOW);
      ledcWrite(CH_Y, pwmDutyY);
    }else{
      digitalWrite(rote_pinY, HIGH);
      ledcWrite(CH_Y, pwmDutyY);
    }
  }
    
  delay(loopDelay);
}


//Timer関連
void ms_timer_st(){
  time_mSt= millis();
}
unsigned long ms_timer_rd(){
  time_ms=millis()-time_mSt;
  return(time_ms);
}

//逆運動学の計算
void ikL(int x,int y){
  float th1, th2, tha, thb, thc, thd,la, lb;
  la=sqrt((x-l/2.0)*(x-l/2.0) + y*y);
  lb=sqrt((x+l/2.0)*(x+l/2.0) + y*y);
  
  tha=atan2(y,(x-l/2.0));
  thb = acos((L1*L1 + la*la - L2*L2)/(2*L1*la));  
  thc=PI-atan2(y,(x+l/2.0));
  thd= acos((L1*L1 + lb*lb - L2*L2)/(2*L1*lb));  
  
  th1 = (tha-thb) * 180.0 / PI;
  th2 = (thc-thd) * 180.0 / PI;

  Serial.print(th1);
  Serial.print(", ");
  Serial.println(th2);

  servoValL1 = map(th1, 90.0, 0.0 , 3110.0, 5230.0);
  servoValL2 = map(th2, 0.0, 90.0 , 3170.0, 5270.0);

  ledcWrite(CH_servoL1, int(servoValL1));
  ledcWrite(CH_servoL2, int(servoValL2));
}

void ikR(int x,int y){
  float th1, th2, tha, thb, thc, thd,la, lb;
  la=sqrt((x-l/2.0)*(x-l/2.0) + y*y);
  lb=sqrt((x+l/2.0)*(x+l/2.0) + y*y);
  
  tha=atan2(y,(x-l/2.0));
  thb = acos((L1*L1 + la*la - L2*L2)/(2*L1*la));  
  thc=PI-atan2(y,(x+l/2.0));
  thd= acos((L1*L1 + lb*lb - L2*L2)/(2*L1*lb));  
  
  th1 = (tha-thb) * 180.0 / PI;
  th2 = (thc-thd) * 180.0 / PI;

  Serial.print(th1);
  Serial.print(", ");
  Serial.println(th2);

  servoValR1 = map(th1, 0.0, 90.0 , 3290.0, 5400.0);
  servoValR2 = map(th2, 90.0, 0.0 , 3180.0, 5290.0);

  ledcWrite(CH_servoR1, int(servoValR1));
  ledcWrite(CH_servoR2, int(servoValR2));
}
