#include <WiFi.h>
#include <WebServer.h>
#include <Preferences.h>
#include "driver/pcnt.h"

WebServer server(80);

const char ssid[] = "ServoSet";  // SSID
const char pass[] = "password";   // password

const IPAddress ip(192, 168, 22, 44);      // IPアドレス
const IPAddress subnet(255, 255, 255, 0); // サブネットマスク


#define servoL1 27
#define servoL2 14
#define servoR1 4
#define servoR2 13

#define CH_servoL1 14
#define CH_servoL2 15
#define CH_servoR1 12
#define CH_servoR2 13

float servoValL1, servoValL2, servoValR1, servoValR2;

int angle = 0;
int Dmin = 3000;
int Dmax = 5500;

//ブラウザ表示
void handleRoot() {
  
  String temp ="<!DOCTYPE html> \n<html lang=\"ja\">";
  temp +="<head>";
  temp +="<meta charset=\"utf-8\">";
  temp +="<title>Servo Settting</title>";
  temp +="<meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">";
  temp +="<style>";
  temp +=".container{";
  temp +="  max-width: 500px;";
  temp +="  margin: auto;";
  temp +="  text-align: center;";
  temp +="  font-size: 1.2rem;";
  temp +="}";
  temp +="span,.pm{";
  temp +="  display: inline-block;";
  temp +="  border: 1px solid #ccc;";
  temp +="  width: 50px;";
  temp +="  height: 30px;";
  temp +="  vertical-align: middle;";
  temp +="  margin-bottom: 20px;";
  temp +="}";
  temp +="span{";
  temp +="  width: 120px;";
  temp +="}";
  temp +="button{";
  temp +="  width: 100px;";
  temp +="  height: 40px;";
  temp +="  font-weight: bold;";
  temp +="  margin-bottom: 20px;";
  temp +="}";
  temp +="button.on{ background:lime; color:white; }";
  temp +=".column-2{ max-width:250px; margin:auto; text-align:center; display:flex; justify-content:space-between; flex-wrap:wrap; }";
  temp +="</style>";
  temp +="</head>";
  
  temp +="<body>";
  temp +="<div class=\"container\">";
  temp +="<h3>Servo Settting</h3>";


  //angle
  temp +="サーボ角度<br>";
  temp +="<a class=\"pm\" href=\"/angleM\">-</a>";
  temp +="<span>" + String(angle) + "</span>";
  temp +="<a class=\"pm\" href=\"/angleP\">+</a><br>";

  //Dmin
  temp +="Duty min.<br>";
  temp +="<a class=\"pm\" href=\"/DminM\">-</a>";
  temp +="<span>" + String(Dmin) + "</span>";
  temp +="<a class=\"pm\" href=\"/DminP\">+</a><br>";

  //Dmax
  temp +="Duty max.<br>";
  temp +="<a class=\"pm\" href=\"/DmaxM\">-</a>";
  temp +="<span>" + String(Dmax) + "</span>";
  temp +="<a class=\"pm\" href=\"/DmaxP\">+</a><br>";
  
   
  temp +="</div>";
  temp +="</body>";
  server.send(200, "text/HTML", temp);

}


void handleAngleM() {
  if(angle > 0){
    angle -= 10;
  }
  handleRoot();
}
void handleAngleP() {
  if(angle <= 90){
    angle += 10;
  }
  handleRoot();
}

void handleDminM() {
  if(Dmin > 1500){
    Dmin -= 10;
  }
  handleRoot();
}
void handleDminP() {
  if(Dmin <= 9000){
    Dmin += 10;
  }
  handleRoot();
}

void handleDmaxM() {
  if(Dmax > 1500){
    Dmax -= 10;
  }
  handleRoot();
}
void handleDmaxP() {
  if(Dmax <= 9000){
    Dmax += 10;
  }
  handleRoot();
}


//Core0
void disp(void *pvParameters) {
  for (;;){
    server.handleClient();
    disableCore0WDT();
  }
}


void setup() {
  Serial.begin(115200);

  ledcSetup(CH_servoL1, 50, 16);
  ledcAttachPin(servoL1, CH_servoL1);
  ledcSetup(CH_servoL2, 50, 16);
  ledcAttachPin(servoL2, CH_servoL2);
  ledcSetup(CH_servoR1, 50, 16);
  ledcAttachPin(servoR1, CH_servoR1);
  ledcSetup(CH_servoR2, 50, 16);
  ledcAttachPin(servoR2, CH_servoR2);

  delay(500); 

  
  WiFi.softAP(ssid, pass);           // SSIDとパスの設定
  delay(100);                        // 追記：このdelayを入れないと失敗する場合がある
  WiFi.softAPConfig(ip, ip, subnet); // IPアドレス、ゲートウェイ、サブネットマスクの設定
  
  IPAddress myIP = WiFi.softAPIP();  // WiFi.softAPIP()でWiFi起動

  server.on("/", handleRoot); 
  

  server.on("/angleM", handleAngleM);
  server.on("/angleP", handleAngleP);
  server.on("/DminM", handleDminM);
  server.on("/DminP", handleDminP);
  server.on("/DmaxM", handleDmaxM);
  server.on("/DmaxP", handleDmaxP);

  server.begin();
  
  xTaskCreatePinnedToCore(
    disp
    ,  "disp"   // A name just for humans
    ,  4096  // This stack size can be checked & adjusted by reading the Stack Highwater
    ,  NULL
    ,  1  // Priority, with 3 (configMAX_PRIORITIES - 1) being the highest, and 0 being the lowest.
    ,  NULL 
    ,  0);
   
}


void loop() {
  servoValL1 = map(angle, 90.0, 0.0 , float(Dmin), float(Dmax));
  servoValL2 = map(angle, 0.0, 90.0 , float(Dmin), float(Dmax));
  servoValR1 = map(angle, 0.0, 90.0 , float(Dmin), float(Dmax));
  servoValR2 = map(angle, 90.0, 0.0 , float(Dmin), float(Dmax));

  ledcWrite(CH_servoL1, int(servoValL1));
  ledcWrite(CH_servoL2, int(servoValL2));
  ledcWrite(CH_servoR1, int(servoValR1));
  ledcWrite(CH_servoR2, int(servoValR2));
}
